import logging

from kensu.utils.helpers import singleton


@singleton
class KensuProvider(object):
    default = None  # type: Kensu

    def instance(self):
        # type: () -> Kensu
        return self.default

    def setKensu(self, kensu):
        self.default = kensu

    # FIXME: probably we don't need get_explicit_code_version_fn & get_code_version_fn anymore. but keeping it for backwards-compat for now...
    @staticmethod
    def initKensu(kensu_ingestion_url=None, kensu_ingestion_token=None, process_name=None, user_name=None, code_location=None,
                  get_code_version_fn=None, get_explicit_code_version_fn=None, do_report=None,
                  report_to_file=None, offline_file_name=None, reporter=None, report_process_info=True, **kwargs):
        allow_reinit = kwargs["allow_reinit"] if "allow_reinit" in kwargs else False
        ksu_provided_inst = KensuProvider().instance()
        if ksu_provided_inst is not None and allow_reinit:
            logging.warning("KensuProvider.initKensu called more than once - reinitializing as requested by allow_reinit=True")
            KensuProvider().setKensu(None)
        if ksu_provided_inst is None or allow_reinit:
            from kensu.utils.kensu import Kensu
            pandas_support = kwargs.get("pandas_support", True)

            sklearn_support = kwargs.get("sklearn_support")

            tensorflow_support = kwargs.get("tensorflow_support")

            bigquery_support = kwargs.get("bigquery_support")
            bigquery_headers = kwargs.get("bigquery_headers")

            project_name = kwargs.get("project_name")
            environment = kwargs.get("environment")
            execution_timestamp = kwargs.get("execution_timestamp")
            logical_data_source_naming_strategy = kwargs.get("logical_data_source_naming_strategy")
            get_code_version = kwargs.get("get_code_version")
            stats = kwargs.get("compute_stats")
            input_stats = kwargs.get("compute_input_stats")
            compute_delta = kwargs.get("compute_delta_stats")
            kensu_sql_parser_url = kwargs.get("kensu_sql_parser_url")
            api_verify_ssl = kwargs.get("kensu_api_verify_ssl")

            _kensu = Kensu(kensu_ingestion_url=kensu_ingestion_url, kensu_ingestion_token=kensu_ingestion_token, process_name=process_name,
                           user_name=user_name, code_location=code_location, kensu_api_verify_ssl=api_verify_ssl,
                           do_report=do_report, pandas_support=pandas_support, sklearn_support=sklearn_support,
                           bigquery_support=bigquery_support, tensorflow_support=tensorflow_support,
                           project_name=project_name, environment=environment, execution_timestamp=execution_timestamp,
                           logical_data_source_naming_strategy=logical_data_source_naming_strategy,
                           report_to_file=report_to_file, offline_file_name=offline_file_name, reporter=reporter,
                           get_code_version=get_explicit_code_version_fn or get_code_version or get_code_version_fn,
                           compute_stats=stats, compute_input_stats=input_stats, bigquery_headers=bigquery_headers,
                           kensu_sql_parser_url=kensu_sql_parser_url, compute_delta_stats=compute_delta,
                           report_process_info=report_process_info)

            KensuProvider().setKensu(_kensu)
            return _kensu
        else:
            logging.warning("Kensu default is already set kensu={}".format(str(ksu_provided_inst)))



import logging

import boto3
from boto3 import *

from kensu.client import DataSourcePK, DataSource, FieldDef, SchemaPK, Schema
from kensu.requests.models import ksu_str
from kensu.itertools import kensu_list
from kensu.utils.kensu_provider import KensuProvider
from kensu.utils.helpers import logical_naming_batch

class ksu_dict(dict):
    ksu_metadata = {}

    @property
    def __class__(self):
        return dict

    def __getitem__(self, item):
        result = super(ksu_dict, self).__getitem__(item)
        if isinstance(result,str):
            ksu_result = ksu_str(result)
        elif isinstance(result,list):
            ksu_result = kensu_list(result)
        elif isinstance(result,dict):
            ksu_result = ksu_dict(result)
        else:
            ksu_result = result

        if hasattr(ksu_result,'ksu_metadata'):
            ksu_result.ksu_metadata = self.ksu_metadata
        return ksu_result


def create_timestream_ds_schema(db, table, schema_dict):
    from kensu.utils.dsl.extractors.external_lineage_dtos import KensuDatasourceAndSchema
    ksu = KensuProvider().instance()
    ds_name = fmt_timestream_name(db, table)
    return KensuDatasourceAndSchema.for_path_with_opt_schema(
        ksu=ksu,
        ds_path=fmt_timestream_uri(db, table),
        format="TimeStream table",
        categories=fmt_timesteam_lds_names(db,table),
        maybe_schema=schema_dict.items(),
        ds_name=ds_name,
    )

def kensu_put(event_params, event_ctx, **kwargs):
    if isinstance(event_params.get('Body'), ksu_str):
        kensu = KensuProvider().instance()
        put_body = event_params.get('Body')

        #The input is the ksu_str, the metadata contains its schema pk
        input_schema = put_body.metadata['real_schema']
        short_schema = put_body.metadata['short_schema']

        input_location = put_body.metadata['ds_location']
        s3_bucket = event_params.get('Bucket') or 'unknown-s3-bucket'
        s3_key = event_params.get('Key') or 'unknown-s3-key'

        # Creation of the output datasource (stored in S3)
        location = 'aws::S3::' + s3_bucket + '/' + s3_key
        name = s3_key

        if kensu.logical_naming == 'ReplaceNumbers':
            logical = logical_naming_batch(name)
        else:
            logical = name

        result_pk = DataSourcePK(location=location,
                                 physical_location_ref=kensu.default_physical_location_ref)
        result_ds = DataSource(name=name, categories=['logical::'+logical],format=name.split('.')[-1],
                               pk=result_pk)._report()

        input_fields = [k.name for k in short_schema.pk.fields]
        input_schema_pk = short_schema.to_guid()

        # This data source has the same schema fields as the input

        fields = [FieldDef(name=k.name, field_type=k.field_type, nullable=True) for k in input_schema.pk.fields]

        sc_pk = SchemaPK(result_ds.to_ref(),
                         fields=fields)
        result_sc = Schema(name="schema:" + result_ds.name, pk=sc_pk)._report()

        fields = [FieldDef(name=k.name, field_type=k.field_type, nullable=True) for k in short_schema.pk.fields]

        sc_pk = SchemaPK(result_ds.to_ref(),
                         fields=fields)

        short_result_sc = Schema(name="short-schema:" + result_ds.name, pk=sc_pk)._report()

        kensu.real_schema_df[short_result_sc.to_guid()] = put_body.metadata['stats']

        for col in input_fields:
            kensu.add_dependencies_mapping(short_result_sc.to_guid(),str(col),input_schema_pk,str(col),'s3_put')
        kensu.report_with_mapping()


def fmt_timestream_uri(db, table):
    return f"aws::TimeStream://{db}.{table}"


def fmt_timestream_name(db, table):
    return f"TimeStream://{db}.{table}"


def fmt_timestream_lds_name(db, table):
    return fmt_timestream_name(db, table)


def fmt_timesteam_lds_names(db, table):
    # FIXME: for now DS name and LDS name (and LDS location) must match (need to update write_reconds otherwise)
    return ['logical::' + fmt_timestream_name(db, table),
            # 'logicalLocation::' + fmt_timestream_uri(db, table),
            ]

def kensu_timestream_query(event_params):
    logging.info('KENSU: handler of timestream_query.query...')
    # if true, we'll report all columns found in the table, doesn't matter what columns were actually referred
    # FIXME: do we care about exact columns which were referred?
    ALWAYS_REPORT_ALL_COLUMNS = True
    from pprint import pprint

    sql = event_params['QueryString']
    # ignore kensu agent issued queries
    if 'DESCRIBE ' in sql:
        return

    # FIXME: pass custom config if any - context['client_config'] / context['client_region']
    client = boto3.client('timestream-query')
    aws_lineage = client.prepare_query(
        QueryString=sql,
        ValidateOnly=True  # do not store the query in the DB
    )
    # FIXME: also maybe need to check the query type (is it SELECT, INSERT etc?)

    # P.S. "Aliased" property is "lying" in case we're querying a subquery, e.g.
    # `select * from (select truck_id as truck_id_renamed from tbl)` will say Aliased=False,
    # while in fact the column was renamed and we don't know it's original name
    # so we need to verify existence of columns
    resolved_db_tables = set([
        (c['DatabaseName'], c['TableName'])
        for c in aws_lineage.get('Columns', [])
        if c.get('DatabaseName') and c.get('TableName')
    ])
    # add tables from sqlineage which might have been unresolved by `client.prepare_query`
    try:
        from sqllineage.runner import LineageRunner
        lin_result = LineageRunner(sql, verbose=True)
        maybe_source_tables = [
            (t.schema.raw_name, t.raw_name)
            for t in lin_result.source_tables
            if t.schema and not (t.schema.raw_name, t.raw_name) in resolved_db_tables]
        # sqllineage knows nothing about actual database, so check if such tables exist
        for (db, table) in maybe_source_tables:
            exists = True
            try:
                client.query(QueryString=f'DESCRIBE "{db}"."{table}"')
            except:
                exists = False
            if exists:
                resolved_db_tables.add((db, table))

            # col_lin = lin_result.get_column_lineage(exclude_subquery=False)
    except:
        logging.info(f"Exception while extracting sqllineage for timesteam-query")


    # FIXME: in some cases even the table name is not available in prepare_query result, use sqllineage as fallback
    # e.g. when refering to a column of certain table only inside expression
    full_table_schemas = {
        (db, table): {row['Data'][0]['ScalarValue']: row['Data'][1]['ScalarValue']
                       for row in client.query(QueryString=f'DESCRIBE "{db}"."{table}"').get('Rows', [])}
        for (db, table) in resolved_db_tables
    }
    # not renamed columns, if any
    unchecked_known_input_columns = [
        (c['Name'], c['DatabaseName'], c['TableName'])
        # {'column_name': c['Name'],
        #  'db': c['DatabaseName'],
        #  'table': c['TableName']}
        for c in aws_lineage.get('Columns', [])
        if c.get('DatabaseName') and c.get('TableName') and c.get('Name')
    ]
    validated_discovered_input_columns = [
        (db, table, col_name, full_table_schemas.get((db, table), {}).get(col_name))
        for (col_name, db, table) in unchecked_known_input_columns
        if full_table_schemas.get((db, table), {}).get(col_name)
    ]
    from collections import defaultdict
    validated_input_columns = defaultdict(lambda: {})
    for (db, table, col_name, data_type) in validated_discovered_input_columns:
        validated_input_columns[(db, table)][col_name] = data_type

    # FIXME: get input columns from sqllineage too...
    # for each discovered input table, if we didn't find any input columns, assume that all columns were accessed
    for (db, table) in resolved_db_tables:
        if not validated_input_columns.get((db, table)) or ALWAYS_REPORT_ALL_COLUMNS:
            logging.warning(f"Kensu was unable to find exact input columns for timestream query table '{db}'.'{table}', using all columns...")
            validated_input_columns[(db, table)] = full_table_schemas.get((db, table), {})

    #print("Kensu Timestream-query:")
    #pprint(resolved_db_tables)
    #pprint(full_table_schemas)
    #pprint(validated_discovered_input_columns)
    #print("final:")
    #pprint(validated_input_columns)
    # report read operation to Kensu
    from kensu.utils.dsl.extractors.external_lineage_dtos import KensuDatasourceAndSchema
    ksu = KensuProvider().instance()
    for ((db, table), schema_dict) in validated_input_columns.items():
        ds = create_timestream_ds_schema(db, table, schema_dict)
        if ksu.lean_mode:
            ds.ksu_ds._report()
            ds.ksu_schema._report()
            ksu.real_schema_df[ds.ksu_schema.to_guid()] = None  # is this one needed?
            ksu.register_input_lean_mode(ds.ksu_schema)
        else:
            logging.warning("Timestream-query.query() kensu tracking is supported only in kensu_lean_mode=True")


def dim_to_schema(dim):
    return (dim.get('Name'), dim.get('DimensionValueType', 'unknown'))


# FIXME: infer datatypes for measures and dimensions

def dimensions_schema(rec):
    return [dim_to_schema(d)
            for d in rec.get('Dimensions', [])]
# e.g.:
# ---
#     'MeasureName': 'IoTMulti-stats',
#     'MeasureValues': [
#             {'Name': 'load', 'Value': '12.3', 'Type': 'DOUBLE'},
#             #{'Name': 'fuel-reading	', 'Value': '13.4', 'Type': 'DOUBLE'},
#         ],
#     'MeasureValueType': 'MULTI',
def measures_schema(rec):
    n = rec.get('MeasureName')
    data_type = rec.get('MeasureValueType', 'DOUBLE')
    # FIXME: special case if MULTI
    if not n:
        return []
    elif data_type == 'MULTI':
        values = rec.get('MeasureValues', [])
        return [(v.get('Name'), v.get('Type'))
                for v in values]
    else:
        return [(n, data_type)]


# returns a dict: {fieldName -> fieldType}
def extract_timestream_write_schema(records, common_attributes):
    # FIXME: take into account CommonAttributes
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/timestream-write.html#TimestreamWrite.Client.write_records
    if len(records) == 0:
        return {}
    else:
        rec = records[0]
        schema = dimensions_schema(rec) + measures_schema(rec) + [('time', 'timestamp')]
        if common_attributes:
            schema = schema + dimensions_schema(common_attributes) + measures_schema(common_attributes)
        return dict([(n.lower(), data_type.lower())
                     for (n, data_type) in schema])

def kensu_write_records(event_params):
    database = event_params['DatabaseName']
    table = event_params['TableName']
    records = list(event_params['Records'])

    kensu = KensuProvider().instance()
    # Creation of the output datasource (stored in S3)
    schema_dict = extract_timestream_write_schema(records, event_params.get('CommonAttributes'))
    ds = create_timestream_ds_schema(database, table, schema_dict)
    ds.ksu_ds._report()
    schema = ds.ksu_schema._report()

    # TODO: Create Stats

    if kensu.lean_mode:
        kensu.outputs_lean.append(schema)
        if ds.ksu_ds.name in kensu.ds_name_stats:
            stats_json = kensu.ds_name_stats[ds.ksu_ds.name]
        else:
            stats_json = None
        kensu.real_schema_df[schema.to_guid()]= stats_json
        kensu.report_without_mapping()




def add_custom_method(class_attributes, **kwargs):
    class_attributes['kensu_put'] = kensu_put

    original_get = class_attributes['get']

    def kensu_get(event_params, **kwargs):
        result = original_get(event_params,**kwargs)

        if 'Body' in result:
            result = ksu_dict(result)

            s3_bucket = event_params.bucket_name or 'unknown-s3-bucket'
            s3_key = event_params.key or 'unknown-s3-key'

            # Creation of the output datasource (stored in S3)
            location = 'aws::S3::' + s3_bucket + '/' + s3_key
            name = s3_key
            result.ksu_metadata['origin_location'] = location
            result.ksu_metadata['origin_name'] = name

            import botocore.response as btr
            from kensu.botocore.response import StreamingBody

            if isinstance(result['Body'],btr.StreamingBody):
                result['Body'].__class__ = StreamingBody
                result['Body'].ksu_metadata = result.ksu_metadata
        return result

    class_attributes['get'] = kensu_get

boto3._get_default_session().events.register("creating-resource-class.s3.Object",
                        add_custom_method)



def kensu_tracker(*class_attributes, **kwargs):
    import logging
    param_types = [
    ]
    import pprint
    if kwargs.get('params'):
        param_types = [[k, v, type(v)] for k, v in kwargs.get('params').items()]
    logging.debug('---\nKensu AWS tracker: '
          'param_types: {}\n'
          'class_attributes:{}\n kwargs: {}\n-----'.format(
        str(param_types),
        str(class_attributes),
        pprint.pformat(kwargs) + '\n'+ str([ [k, v, type(v)] for k,v in kwargs.items()])))

    event_name = kwargs.get('event_name')
    event_params = kwargs.get('params')
    event_ctx = kwargs.get('context')
    if event_name == 'provide-client-params.s3.PutObject' and event_params:
        kensu_put(event_params=event_params, event_ctx=event_ctx, **kwargs)
    if event_name == 'provide-client-params.timestream-write.WriteRecords' and event_params:
        kensu_write_records(event_params)
    if event_name == 'provide-client-params.timestream-query.Query' and event_params:
        kensu_timestream_query(event_params)

#boto3._get_default_session().events.register('creating-resource-class.s3.ServiceResource',kensu_tracker)
#boto3._get_default_session().events.register('before-send.s3.PutObject', kensu_tracker)
boto3._get_default_session().events.register('provide-client-params.s3.PutObject', kensu_tracker)

# in case we wanted to see all events - use *
#S3 = boto3.resource('s3' , region_name='eu-west-1')
#event_system = S3.meta.client.meta.events
#event_system.register("*",kensu_tracker)
#event_system.register('creating-resource-class.s3.*', prt)

#boto3._get_default_session().events.register("*",kensu_tracker)

boto3._get_default_session().events.register('provide-client-params.timestream-write.WriteRecords', kensu_tracker)
boto3._get_default_session().events.register('provide-client-params.timestream-query.Query', kensu_tracker)



from kensu.utils.dsl.process_lineage_deps_builder import ProcessLineageDepsBuilder


class LineageBuilder(object):
  def __init__(self, kensu, report_stats):
    self.kensu = kensu
    self.deps = []
    self.report_stats = report_stats

  @property
  def n(self):
    return self.new_dependency()
  def new_dependency(self):
    return ProcessLineageDepsBuilder(self.kensu, self)
  def add_deps(self, dep):
    self.deps.append(dep)

  def e(self, **kwargs):
    return self.end_lineage(**kwargs)
  def end_lineage(self, **kwargs):
    self.kensu.new_lineage(self.deps, self.report_stats, **kwargs)



import logging
from hashlib import sha256

import matplotlib.pyplot as plt


import kensu
from kensu.client import *
from kensu.utils.dsl.extractors import ExtractorSupport, get_or_set_rand_location
from kensu.utils.helpers import singleton, save_stats_json, to_datasource
from kensu.utils.kensu_provider import KensuProvider

@singleton
class PlotSupport(ExtractorSupport):  # should extends some KensuSupport class

    def is_supporting(self, fig):
        return isinstance(fig,  plt.Figure)

    def is_machine_learning(self, df):
        return False

    # return list of FieldDef
    def extract_schema_fields(self, fig):
        k=KensuProvider().instance()
        fields = []
        for ax in fig.axes:
            for df in ax.inheritance:
                fields_ax_df = k.extractors.extract_schema_fields(df)
                fields = fields + fields_ax_df
        return fields

    # return dict of doubles (stats)
    def extract_stats(self, fig):
        kensu = KensuProvider().instance()
        stats = {}
        ax_num = 0
        for ax in fig.axes:
            source_num = 0
            for df in ax.inheritance:
                #TODO : extract ax name if any instead of generic number
                inherited_stats = {'Ax'+str(ax_num)+'.Source'+str(source_num)+'.'+key : value for key,value in kensu.extractors.extract_stats(df).items()}
                stats = {**stats,**inherited_stats}
                source_num+=1
            ax_num += 1
        return stats

    def extract_data_source(self, fig, pl, **kwargs):

        logical_data_source_naming_strategy = kwargs["logical_data_source_naming_strategy"] if "logical_data_source_naming_strategy" in kwargs else None
        location = kwargs["location"] if "location" in kwargs else None
        fmt = 'matplotlib image'

        if location is None or fmt is None:
            raise Exception(
                "cannot report new matplotlib data source without location ({}) a format provided ({})!".format(location,
                                                                                                          fmt))

        ds_pk = DataSourcePK(location=location, physical_location_ref=pl)


        name = ('/').join(location.split('/')[-2:])

        ds=to_datasource(ds_pk, fmt, location, logical_data_source_naming_strategy, name)
        return ds

    def extract_schema(self, data_source, fig):
        fields = self.extract_schema_fields(fig)
        sc_pk = SchemaPK(data_source.to_ref(), fields=fields)
        schema = Schema(name="schema:" + data_source.name, pk=sc_pk)
        return schema

    def extract_data_source_and_schema(self, df, pl, **kwargs):
        df = self.skip_wr(df)
        ds = self.extract_data_source(df, pl, **kwargs)
        sc = self.extract_schema(ds, df)
        return ds, sc


from kensu.utils.kensu_provider import KensuProvider
import warnings
import logging



def add_rule(data_source, field, type, parameters,context="DATA_STATS"):
    k = KensuProvider().instance()
    k.rules.append({data_source:
                        {'field': field,
                         'fun': {"name": type, "arguments": parameters},
                         'context':context
                         }})


def add_not_null_rule(
        ksu,  # type: Kensu
        lds_name,
        non_null_col,
        null_suffix='nullrows'
):
    try:
        logging.info(f"KENSU: Adding a Kensu rule: NOT_NULL({non_null_col}) on LDS={lds_name}")
        lds_context = "LOGICAL_DATA_SOURCE"
        add_rule(data_source=lds_name,
                 field=f'{non_null_col}.{null_suffix}',
                 type='Range',
                 parameters={'maxVal': 0},
                 context=lds_context)
        ksu.send_rules()
    except Exception as e:
        logging.warning(f"Error while adding a Kensu rule  NOT_NULL({non_null_col}.{null_suffix})", e)


def add_min_max(data_source, field, min = None, max = None, context="DATA_STATS"):
    parameters = {}
    if min is not None:
        parameters["minVal"] = min
    if max is not None:
        parameters["maxVal"] = max
    add_rule(data_source,field,type='Range',parameters=parameters,context=context)


def add_missing_value_rules(data_source, data_frame=None, cols=None, context="DATA_STATS"):
    if cols is None:
        cols = data_frame.columns
    for col in cols:
        field = col+'.nullrows'
        add_rule(data_source, field, type='Range', parameters={'maxVal':0}, context=context)


def add_frequency_rule(data_source, hours=None, days=None, weeks=None, months=None):
    if hours:
        timeLapseUnit = 'Hours'
        timeLapse = hours
    elif days:
        timeLapseUnit = 'Days'
        timeLapse = days
    elif weeks:
        timeLapseUnit = 'Weeks'
        timeLapse = weeks
    elif months:
        timeLapseUnit = 'Months'
        timeLapse = months
    else:
        logging.info("KENSU: add_frequency_rule failed: no time unit specified")
        return None

    add_rule(data_source,
             field=None,
             type='Frequency',
             parameters={'timeLapse': timeLapse,'timeLapseUnit':timeLapseUnit})


#TODO Variabilty over time
def add_variability_rule(data_source, field, variation_in_percent, hours = None, days = None, weeks = None, months = None,context = "DATA_STATS"):
    parameters = {}
    parameters['variation'] = variation_in_percent

    add_rule(data_source,field=field,
             type='Variability',
             parameters=parameters,context=context)


def add_variability_constraint_data_source(data_source, field, variation_in_percent, hours = None, days = None, weeks = None, months = None):
    context = "LOGICAL_DATA_SOURCE"
    return add_variability_rule(data_source, field, variation_in_percent, hours, days, weeks, months,context)


def check_format_consistency(data_source):
    k_sdk = KensuProvider().instance().sdk
    checked_format = k_sdk.get_latest_datasource_in_logical(data_source)['format']
    previous_ds = k_sdk.get_latest_datasource_in_logical(data_source, n=-2)

    if checked_format and previous_ds:
        previous_format = previous_ds['format']
        bool = checked_format == previous_format
        if bool:
            pass
        else:
            logging.warning("KENSU: The format of the datasource {} is not consistent, expected {}, got {}".format(data_source,previous_format, checked_format))


def check_schema_consistency(data_source):
    k_sdk = KensuProvider().instance().sdk
    checked_schema = k_sdk.get_latest_schema_in_logical(data_source, n=-1)
    previous_schema = k_sdk.get_latest_schema_in_logical(data_source, n=-2)

    if checked_schema and previous_schema:
        # Check of field changes
        missing_keys = previous_schema.keys() - checked_schema.keys()
        if missing_keys:
            logging.warning("KENSU: The following key(s) are missing from {} : {}".format(data_source,list(missing_keys)))

        new_keys =  checked_schema.keys() - previous_schema.keys()
        if new_keys:
            logging.warning("KENSU: The following key(s) are new in {} : {}".format(data_source,list(new_keys)))

        all_changed_fields = list(missing_keys)+list(new_keys)

        type_diff = dict(set(checked_schema.items())-set(previous_schema.items()))
        for x in all_changed_fields:
            if x in type_diff:
                del type_diff[x]
        if type_diff:
            for y in type_diff.keys():
                logging.warning("KENSU: The following field in {} has a wrong type: {} Expected {}, got {}".format(data_source,y,previous_schema[y],checked_schema[y]))


# TODO WIP check nrows
def check_nrows_consistency(how = "minimum"):
    k = KensuProvider().instance()
    k.check_rules.append({"nrows_consistency":how})


def create_kensu_nrows_consistency(how):
    k = KensuProvider().instance()
    for lineage in k.lineage_and_ds:

        inputs_nrows= [{k.logical_name_by_guid[schema]: k.schema_stats[schema]['nrows']}
                        for schema in k.lineage_and_ds[lineage]['from_schema_ref']
                        if 'nrows' in k.schema_stats[schema]]

        values = []
        for dic in inputs_nrows:
            for key in dic:
                values.append(dic[key])

        output_name = k.logical_name_by_guid[list(k.lineage_and_ds[lineage]['to_schema_ref'])[0]]
        checked_rules = {output_name : []}
        if how in ['minimum','maximum']:
            if how == "minimum":
                min_value = min(values)
            elif how == "maximum":
                min_value = max(values)
            ds_in_candidates = []
            for val in inputs_nrows:
                key = list(val.keys())[0]
                if val[key] == min_value:
                    ds_in_candidates.append(key)
            for ds in ds_in_candidates:
                field = 'delta.nrows_'+ds.replace('.','_')+'.abs'
                add_min_max(output_name, field, max = 0)
                checked_rules[output_name].append({field:{"output_nrows" : (k.schema_stats[list(k.lineage_and_ds[lineage]['to_schema_ref'])[0]])['nrows'], "input_nrows" : min_value}})
            k.check_rules.append({'check_nrows_consistency':checked_rules})



#TODO Support other how types:::

        #     if output_nrows == min_value:
        #         None
        #     elif output_nrows < min_value:
        #         percent = (output_nrows / min_value) * 100
        #         percent = round(percent,2)
        #         print("{} has less rows than expected: {} out of a maximum of {} - {}%".format(output_name,output_nrows,min_value,percent))
        # elif how == "maximum":
        #     max_value = max(values)
        #     if output_nrows == max_value:
        #         None
        #     elif output_nrows < max_value:
        #         percent = (output_nrows / max_value) * 100
        #         percent = round(percent, 2)
        #         print(
        #             "{} has less rows than expected: {} out of a maximum of {} - {}%".format(output_name, output_nrows,
        #                                                                                      max_value, percent))
        # else:
        #     limiting_ds = how
        #     if limiting_ds in inputs_nrows:
        #         limiting_value = inputs_nrows[limiting_ds]
        #         if output_nrows == limiting_value:
        #             None
        #         elif output_nrows < limiting_value:
        #             percent = (output_nrows / limiting_value) * 100
        #             percent = round(percent, 2)
        #             print(
        #                 "{} has less rows than expected: {} out of a maximum of {} - {}%".format(output_name,
        #                                                                                          output_nrows,
        #                                                                                          limiting_value, percent))
        #
        #



import logging
import re
from hashlib import sha1
import os

# fixme: circular import, so need to inline in each fn?
# from kensu.utils.kensu_provider import KensuProvider

from kensu.client import FieldDef, SchemaPK, Schema, DataSource


def to_snake_case(name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def singleton(cls, *args, **kw):
    instances = {}

    def _singleton():
        if cls not in instances:
            instances[cls] = cls(*args, **kw)
        return instances[cls]

    return _singleton


def to_hash_key(o):
    return sha1(str(o).encode()).hexdigest()


def eventually_report_in_mem(o):
    from kensu.utils.kensu_provider import KensuProvider
    kensu = KensuProvider().instance()
    if kensu.report_in_mem:
        o._report()
    return o


def get_absolute_path(path):
    import os
    if path.startswith('gs:/'):
        return path
    for prefix in ["abfss", "/abfss", "dbfs", "/dbfs"]:
        if path.startswith(prefix):
            path = path.replace(prefix, '')
            prefix = prefix.replace('/', '')
            return prefix + ':' + path
    location = 'file:' + str(os.path.abspath(path))
    return location


def maybe_report(o, report):
    from kensu.utils.kensu_provider import KensuProvider
    kensu = KensuProvider().instance()
    if report:
        o._report()
    return o


def extract_ksu_ds_schema(kensu, orig_variable, report=False, register_orig_data=False):
    ds = kensu.extractors.extract_data_source(orig_variable, kensu.default_physical_location_ref, logical_data_source_naming_strategy=kensu.logical_naming)
    schema = kensu.extractors.extract_schema(ds, orig_variable)
    maybe_report(ds, report=report)
    maybe_report(schema, report=report)
    if register_orig_data:
        kensu.real_schema_df[schema.to_guid()] = orig_variable
    return ds, schema


def report_all2all_lineage(in_obj, out_obj, op_type, in_inmem=True, out_inmem=True):
    from kensu.utils.kensu_provider import KensuProvider
    kensu = KensuProvider().instance()
    in_ds, in_schema = extract_ksu_ds_schema(kensu,
                                             in_obj,
                                             report=kensu.report_in_mem or not in_inmem,
                                             register_orig_data=not in_inmem)
    out_ds, out_schema = extract_ksu_ds_schema(kensu,
                                               out_obj,
                                               report=kensu.report_in_mem or not out_inmem,
                                               register_orig_data=not out_inmem)
    logging.debug("in_schema=" + str(in_schema) + "\nout_schema=" + str(out_schema))
    if kensu.mapping:
        for col in out_obj:
            if col in [v.name for v in in_schema.pk.fields]:
                kensu.add_dependencies_mapping(guid=out_schema.to_guid(),
                                               col=str(col),
                                               from_guid=in_schema.to_guid(),
                                               from_col=str(col),
                                               type=op_type)


def report_simple_copy_with_guessed_schema(input_uri,  # type: str
                                           output_absolute_uri,  # type: str
                                           read_schema_from_filename,  # type: str
                                           operation_type,  # type: str
                                           ):
    from kensu.utils.dsl.extractors.external_lineage_dtos import GenericComputedInMemDs
    try:
        maybe_schema = None
        logging.info(f"report_simple_copy_with_guessed_schema input={input_uri} output={output_absolute_uri}")
        if read_schema_from_filename.endswith('.csv'):
            try:
                import kensu.pandas as pd
                # FIXME: how to handle multiple possibilities for CSV separators? e.g. sep=";"
                maybe_pandas_df = pd.read_csv(read_schema_from_filename)
                from kensu.utils.kensu_provider import KensuProvider
                ksu = KensuProvider().instance()
                _, schema = extract_ksu_ds_schema(ksu, maybe_pandas_df, report=False, register_orig_data=False)
                maybe_schema = [(f.name, f.field_type) for f in schema.pk.fields]
            except Exception as ex:
                logging.warning("unable to guess schema of .csv() file", ex)
        GenericComputedInMemDs.report_copy_with_opt_schema(
            src=input_uri,
            dest=output_absolute_uri,
            operation_type=operation_type,
            maybe_schema=maybe_schema
        )
    except Exception as e:
        logging.warning(f"caught exception in dumb_parse_curl", e)


def flatten(d, parent_key='', sep='.'):
    items = []
    if isinstance(d,list):
        for element in d:
            items.extend(flatten(element, parent_key='[]', sep=sep).items())
    else:
        for k, v in d.items():
            new_key = parent_key + sep + k if parent_key else k
            if isinstance(v, dict):
                items.extend(flatten(v, new_key, sep=sep).items())
            elif isinstance(v, list):
                if isinstance(v[0], dict):
                    new_key = new_key + '[]'
                    for i in v:
                        items.extend(flatten(i, new_key, sep=sep).items())
            else:
                items.append((new_key, type(v).__name__))
    return dict(items)

def logical_naming_batch(string):
    from itertools import groupby, chain
    grouped = groupby(string, str.isdigit)
    return ''.join(chain.from_iterable("<number>" if k else g for k,g in grouped))


def to_datasource(ds_pk, format, location, logical_data_source_naming_strategy, name):
    if logical_data_source_naming_strategy == 'File':
        logical_category = location.split('/')[-1]
        ds = DataSource(name=name, format=format, categories=['logical::' + logical_category], pk=ds_pk)

    elif logical_data_source_naming_strategy == 'Folder':
        logical_category = location.split('/')[-2]
        ds = DataSource(name=name, format=format, categories=['logical::' + logical_category], pk=ds_pk)

    elif logical_data_source_naming_strategy == 'AnteFolder':
        logical_category = location.split('/')[-3]
        ds = DataSource(name=name, format=format, categories=['logical::' + logical_category], pk=ds_pk)

    elif logical_data_source_naming_strategy == 'ReplaceNumbers':
        logical_category = logical_naming_batch(name)
        ds = DataSource(name=name, format=format, categories=['logical::' + logical_category], pk=ds_pk)

    elif callable(logical_data_source_naming_strategy):
        # TODO create to_datasource for all extractors - limited to pandas DataFrame and BigQuery for now
        try:
            logical_category = logical_data_source_naming_strategy(location)
        except Exception as e:
            logging.warning("data source logical_data_source_naming_strategy function passed to initKensu or KensuProvider instance"
                            " returned an exception, "
                            "using default data source naming convention. \n {}".format(e))
            logical_category = location.split('/')[-1]
        ds = DataSource(name=name, format=format, categories=['logical::' + logical_category], pk=ds_pk)

    else:
        ds = DataSource(name=name, format=format, categories=[], pk=ds_pk)
    return ds

def save_stats_json(schemaID,stats):
    from kensu.utils.kensu_provider import KensuProvider
    kensu = KensuProvider().instance()
    kensu.schema_stats[schemaID]=stats

def extract_short_json_schema(result, result_ds):
    fields_set = set()
    if isinstance(result, list):
        for element in result:
            if isinstance(element,dict):
                for e in element.keys():
                    fields_set.add(('[].' + str(e), type((element[e])).__name__))
            else:
                fields_set.add(('value', 'unknown'))
    elif isinstance(result, dict):
        for e in result.keys():
            fields_set.add((e, type(result[e]).__name__))
    else:
        fields_set.add(('value', 'unknown'))
    fields = [FieldDef(name=k[0], field_type=k[1], nullable=True) for k in fields_set]

    sc_pk = SchemaPK(result_ds.to_ref(),fields=fields)

    short_result_sc = Schema(name="short-schema:" + result_ds.name, pk=sc_pk)
    return short_result_sc


def extract_config_property(key, default, arg=None, kw=None, conf=None, tpe=None):
    """
    Looks for a property value following this precedence:
      env_var > arg > kwargs > conf > default
    The default value is used to determine the type of the conf value (it can be overridden by tpe).
    The environment variable will be looked up based on the pattern of `KSU_<upper-key>`.
    """
    env_var_key = "KSU_" + key.upper()
    if os.environ.get(env_var_key) is not None:
        env_var = os.environ.get(env_var_key)
        if tpe is not None:
            env_var = tpe(env_var)
        return env_var
    elif arg is not None:
        return arg
    elif key in kw and kw[key] is not None:
        return kw[key]
    elif key in conf and conf.get(key) is not None:
        if default is not None and tpe is None:
            tpe = type(default)
        r = conf.get(key)
        if tpe is list:
            r = r.replace(" ", "").split(",")
        elif tpe is bool:
            r = conf.getboolean(key)
        elif tpe is not None:
            r = tpe(r)
        return r
    else:
        return default


def get_conf_path(default = "conf.ini"):
    return os.environ["KSU_CONF_FILE"] if "KSU_CONF_FILE" in os.environ else default


# coding: utf-8

"""
    

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)

    OpenAPI spec version: beta
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


from pprint import pformat

from six import iteritems


class BatchEntityReport(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """


    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'projects': 'list[BatchProject]',
        'processes': 'list[BatchProcess]',
        'process_runs': 'list[BatchProcessRun]',
        'process_run_stats': 'list[BatchProcessRunStats]',
        'process_lineages': 'list[BatchProcessLineage]',
        'lineage_runs': 'list[BatchLineageRun]',
        'schemas': 'list[BatchSchema]',
        'schema_field_tags': 'list[BatchSchemaFieldTag]',
        'physical_locations': 'list[BatchPhysicalLocation]',
        'data_sources': 'list[BatchDataSource]',
        'code_versions': 'list[BatchCodeVersion]',
        'code_bases': 'list[BatchCodeBase]',
        'users': 'list[BatchUser]',
        'data_stats': 'list[BatchDataStats]',
        'models': 'list[BatchModel]',
        'model_trainings': 'list[BatchModelTraining]',
        'model_metrics': 'list[BatchModelMetrics]'
    }

    attribute_map = {
        'projects': 'projects',
        'processes': 'processes',
        'process_runs': 'processRuns',
        'process_run_stats': 'processRunStats',
        'process_lineages': 'processLineages',
        'lineage_runs': 'lineageRuns',
        'schemas': 'schemas',
        'schema_field_tags': 'schemaFieldTags',
        'physical_locations': 'physicalLocations',
        'data_sources': 'dataSources',
        'code_versions': 'codeVersions',
        'code_bases': 'codeBases',
        'users': 'users',
        'data_stats': 'dataStats',
        'models': 'models',
        'model_trainings': 'modelTrainings',
        'model_metrics': 'modelMetrics'
    }

    def __init__(self, projects=None, processes=None, process_runs=None, process_run_stats=None, process_lineages=None, lineage_runs=None, schemas=None, schema_field_tags=None, physical_locations=None, data_sources=None, code_versions=None, code_bases=None, users=None, data_stats=None, models=None, model_trainings=None, model_metrics=None):
        """
        BatchEntityReport - a model defined in Swagger
        """

        self._projects = None
        self._processes = None
        self._process_runs = None
        self._process_run_stats = None
        self._process_lineages = None
        self._lineage_runs = None
        self._schemas = None
        self._schema_field_tags = None
        self._physical_locations = None
        self._data_sources = None
        self._code_versions = None
        self._code_bases = None
        self._users = None
        self._data_stats = None
        self._models = None
        self._model_trainings = None
        self._model_metrics = None

        if projects is not None:
          self.projects = projects
        if processes is not None:
          self.processes = processes
        if process_runs is not None:
          self.process_runs = process_runs
        if process_run_stats is not None:
          self.process_run_stats = process_run_stats
        if process_lineages is not None:
          self.process_lineages = process_lineages
        if lineage_runs is not None:
          self.lineage_runs = lineage_runs
        if schemas is not None:
          self.schemas = schemas
        if schema_field_tags is not None:
          self.schema_field_tags = schema_field_tags
        if physical_locations is not None:
          self.physical_locations = physical_locations
        if data_sources is not None:
          self.data_sources = data_sources
        if code_versions is not None:
          self.code_versions = code_versions
        if code_bases is not None:
          self.code_bases = code_bases
        if users is not None:
          self.users = users
        if data_stats is not None:
          self.data_stats = data_stats
        if models is not None:
          self.models = models
        if model_trainings is not None:
          self.model_trainings = model_trainings
        if model_metrics is not None:
          self.model_metrics = model_metrics

    @property
    def projects(self):
        """
        Gets the projects of this BatchEntityReport.

        :return: The projects of this BatchEntityReport.
        :rtype: list[BatchProject]
        """
        return self._projects

    @projects.setter
    def projects(self, projects):
        """
        Sets the projects of this BatchEntityReport.

        :param projects: The projects of this BatchEntityReport.
        :type: list[BatchProject]
        """

        self._projects = projects

    @property
    def processes(self):
        """
        Gets the processes of this BatchEntityReport.

        :return: The processes of this BatchEntityReport.
        :rtype: list[BatchProcess]
        """
        return self._processes

    @processes.setter
    def processes(self, processes):
        """
        Sets the processes of this BatchEntityReport.

        :param processes: The processes of this BatchEntityReport.
        :type: list[BatchProcess]
        """

        self._processes = processes

    @property
    def process_runs(self):
        """
        Gets the process_runs of this BatchEntityReport.

        :return: The process_runs of this BatchEntityReport.
        :rtype: list[BatchProcessRun]
        """
        return self._process_runs

    @process_runs.setter
    def process_runs(self, process_runs):
        """
        Sets the process_runs of this BatchEntityReport.

        :param process_runs: The process_runs of this BatchEntityReport.
        :type: list[BatchProcessRun]
        """

        self._process_runs = process_runs

    @property
    def process_run_stats(self):
        """
        Gets the process_run_stats of this BatchEntityReport.

        :return: The process_run_stats of this BatchEntityReport.
        :rtype: list[BatchProcessRunStats]
        """
        return self._process_run_stats

    @process_run_stats.setter
    def process_run_stats(self, process_run_stats):
        """
        Sets the process_run_stats of this BatchEntityReport.

        :param process_run_stats: The process_run_stats of this BatchEntityReport.
        :type: list[BatchProcessRunStats]
        """

        self._process_run_stats = process_run_stats

    @property
    def process_lineages(self):
        """
        Gets the process_lineages of this BatchEntityReport.

        :return: The process_lineages of this BatchEntityReport.
        :rtype: list[BatchProcessLineage]
        """
        return self._process_lineages

    @process_lineages.setter
    def process_lineages(self, process_lineages):
        """
        Sets the process_lineages of this BatchEntityReport.

        :param process_lineages: The process_lineages of this BatchEntityReport.
        :type: list[BatchProcessLineage]
        """

        self._process_lineages = process_lineages

    @property
    def lineage_runs(self):
        """
        Gets the lineage_runs of this BatchEntityReport.

        :return: The lineage_runs of this BatchEntityReport.
        :rtype: list[BatchLineageRun]
        """
        return self._lineage_runs

    @lineage_runs.setter
    def lineage_runs(self, lineage_runs):
        """
        Sets the lineage_runs of this BatchEntityReport.

        :param lineage_runs: The lineage_runs of this BatchEntityReport.
        :type: list[BatchLineageRun]
        """

        self._lineage_runs = lineage_runs

    @property
    def schemas(self):
        """
        Gets the schemas of this BatchEntityReport.

        :return: The schemas of this BatchEntityReport.
        :rtype: list[BatchSchema]
        """
        return self._schemas

    @schemas.setter
    def schemas(self, schemas):
        """
        Sets the schemas of this BatchEntityReport.

        :param schemas: The schemas of this BatchEntityReport.
        :type: list[BatchSchema]
        """

        self._schemas = schemas

    @property
    def schema_field_tags(self):
        """
        Gets the schema_field_tags of this BatchEntityReport.

        :return: The schema_field_tags of this BatchEntityReport.
        :rtype: list[BatchSchemaFieldTag]
        """
        return self._schema_field_tags

    @schema_field_tags.setter
    def schema_field_tags(self, schema_field_tags):
        """
        Sets the schema_field_tags of this BatchEntityReport.

        :param schema_field_tags: The schema_field_tags of this BatchEntityReport.
        :type: list[BatchSchemaFieldTag]
        """

        self._schema_field_tags = schema_field_tags

    @property
    def physical_locations(self):
        """
        Gets the physical_locations of this BatchEntityReport.

        :return: The physical_locations of this BatchEntityReport.
        :rtype: list[BatchPhysicalLocation]
        """
        return self._physical_locations

    @physical_locations.setter
    def physical_locations(self, physical_locations):
        """
        Sets the physical_locations of this BatchEntityReport.

        :param physical_locations: The physical_locations of this BatchEntityReport.
        :type: list[BatchPhysicalLocation]
        """

        self._physical_locations = physical_locations

    @property
    def data_sources(self):
        """
        Gets the data_sources of this BatchEntityReport.

        :return: The data_sources of this BatchEntityReport.
        :rtype: list[BatchDataSource]
        """
        return self._data_sources

    @data_sources.setter
    def data_sources(self, data_sources):
        """
        Sets the data_sources of this BatchEntityReport.

        :param data_sources: The data_sources of this BatchEntityReport.
        :type: list[BatchDataSource]
        """

        self._data_sources = data_sources

    @property
    def code_versions(self):
        """
        Gets the code_versions of this BatchEntityReport.

        :return: The code_versions of this BatchEntityReport.
        :rtype: list[BatchCodeVersion]
        """
        return self._code_versions

    @code_versions.setter
    def code_versions(self, code_versions):
        """
        Sets the code_versions of this BatchEntityReport.

        :param code_versions: The code_versions of this BatchEntityReport.
        :type: list[BatchCodeVersion]
        """

        self._code_versions = code_versions

    @property
    def code_bases(self):
        """
        Gets the code_bases of this BatchEntityReport.

        :return: The code_bases of this BatchEntityReport.
        :rtype: list[BatchCodeBase]
        """
        return self._code_bases

    @code_bases.setter
    def code_bases(self, code_bases):
        """
        Sets the code_bases of this BatchEntityReport.

        :param code_bases: The code_bases of this BatchEntityReport.
        :type: list[BatchCodeBase]
        """

        self._code_bases = code_bases

    @property
    def users(self):
        """
        Gets the users of this BatchEntityReport.

        :return: The users of this BatchEntityReport.
        :rtype: list[BatchUser]
        """
        return self._users

    @users.setter
    def users(self, users):
        """
        Sets the users of this BatchEntityReport.

        :param users: The users of this BatchEntityReport.
        :type: list[BatchUser]
        """

        self._users = users

    @property
    def data_stats(self):
        """
        Gets the data_stats of this BatchEntityReport.

        :return: The data_stats of this BatchEntityReport.
        :rtype: list[BatchDataStats]
        """
        return self._data_stats

    @data_stats.setter
    def data_stats(self, data_stats):
        """
        Sets the data_stats of this BatchEntityReport.

        :param data_stats: The data_stats of this BatchEntityReport.
        :type: list[BatchDataStats]
        """

        self._data_stats = data_stats

    @property
    def models(self):
        """
        Gets the models of this BatchEntityReport.

        :return: The models of this BatchEntityReport.
        :rtype: list[BatchModel]
        """
        return self._models

    @models.setter
    def models(self, models):
        """
        Sets the models of this BatchEntityReport.

        :param models: The models of this BatchEntityReport.
        :type: list[BatchModel]
        """

        self._models = models

    @property
    def model_trainings(self):
        """
        Gets the model_trainings of this BatchEntityReport.

        :return: The model_trainings of this BatchEntityReport.
        :rtype: list[BatchModelTraining]
        """
        return self._model_trainings

    @model_trainings.setter
    def model_trainings(self, model_trainings):
        """
        Sets the model_trainings of this BatchEntityReport.

        :param model_trainings: The model_trainings of this BatchEntityReport.
        :type: list[BatchModelTraining]
        """

        self._model_trainings = model_trainings

    @property
    def model_metrics(self):
        """
        Gets the model_metrics of this BatchEntityReport.

        :return: The model_metrics of this BatchEntityReport.
        :rtype: list[BatchModelMetrics]
        """
        return self._model_metrics

    @model_metrics.setter
    def model_metrics(self, model_metrics):
        """
        Sets the model_metrics of this BatchEntityReport.

        :param model_metrics: The model_metrics of this BatchEntityReport.
        :type: list[BatchModelMetrics]
        """

        self._model_metrics = model_metrics

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        if not isinstance(other, BatchEntityReport):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other



import logging

import urllib3
urllib3.disable_warnings()
from kensu.utils.kensu import Kensu
from kensu.client.models import *
from kensu.utils.helpers import to_datasource

from pyspark.sql import DataFrame
from pyspark.sql import SparkSession
from kensu.pyspark import get_process_run_info, get_inputs_lineage_fn, register_manual_lineage, get_spark_from_df, \
    get_df_with_inmem_tag
from kensu.utils.kensu_provider import KensuProvider


def get_spark_session():
    import gc
    for obj in gc.get_objects():
        if isinstance(obj, SparkSession):
            kensu_spark_session = obj
    return kensu_spark_session or None

import types


def get_schema_fields_from_spark_df(
        spark_df,  # type: DataFrame
):
    return list([
        FieldDef(name=f.name, field_type=f.dataType.simpleString(), nullable=False)
        for f in spark_df.schema.fields
    ])


def register_in_mem_schema(
        ds,  # type: DataSource
        schem_fields,  # type: list[FieldDef]
        in_mem_location,  # type: str
        in_mem_name  # type: str
):
    schema = Schema(in_mem_location, pk=SchemaPK(data_source_ref=DataSourceRef(by_guid=ds.to_guid()),
                                                 fields=schem_fields))._report()
    k = KensuProvider().instance()
    k.name_schema_lineage_dict[in_mem_name] = schema.to_guid()
    return schema

"""
Tag a Spark DataFrame as in-mem datasource, assuming that it will be transformed into python by using say .toPandas().
The result in Kensu will be: [spark input{1,2,..}] --lineage--> [in-mem: name]

Currently only a single .tagInMem() call is allowed on same Spark DataFrame. 
The subsequent calls would be interpreted as individual in-mem data-sources not dependent on the earlier one (due to
how spark works - DataFrame is never materialized before "write"), i.e.:
    `spark_df.tagInMem('name1').filter('col1 = 1').tagInMem('name2')`
would create two UNRELATED in-mem data-sources.
"""
def tagInMem(self,  # type: DataFrame
             name):
    # Get spark session info
    # FIXME: we can get Spark session more reliably/efficiently from `self: DataFrame`,
    #  while databricks have multiple sessions (and not sure about isolation there)!
    # spark = get_spark_session()
    # FIXME: check if self is Spark DataFrame?
    spark_df = self
    spark = get_spark_from_df(spark_df)
    spark_info = get_process_run_info(spark)
    process = spark_info['process_guid']

    # Creation of an in-mem data source
    spark_ds_name=f'{process}/{name}'
    # location = f'in-memory-data://{process}/{name}'
    location = f'in-memory-data://{spark_ds_name}'
    format = 'in-mem'
    k = KensuProvider().instance()
    ds_pk = DataSourcePK(location=location,
                         physical_location_ref=PhysicalLocationRef(by_pk=k.UNKNOWN_PHYSICAL_LOCATION.pk))
    df_ds = to_datasource(ds_pk, format, location, 'File', name)._report()
    spark_df_schema = register_in_mem_schema(
        ds=df_ds,
        schem_fields=get_schema_fields_from_spark_df(spark_df),
        in_mem_location=location,
        in_mem_name=name
    )

    # report Spark lineage to Kensu
    # FIXME: check if self is Spark DataFrame
    # report lineage to Kensu
    spark_lineage = get_inputs_lineage_fn(kensu_instance=k, df=spark_df, virtual_ds_name=spark_ds_name)
    print('got lineage from spark, just before calling .toPandas()')
    print(spark_lineage)
    # FIXME: we could use this and also report spark datastats
    # FIXME: or should we report using low level apis instead, so we'd easily keep the process info extracted from spark
    # spark_lineage.report(
    #     ksu=k,
    #     df_result=spark_df_schema,  # we don't handle Spark DataFrame directly in extractors yet, so pass a schema
    #     operation_type='Spark to Python',
    #     report_output=True,
    #     register_output_orig_data=False
    # )
    # k.report_with_mapping()
    report_spark_lineage(spark_lineage, k=k, result_schema=spark_df_schema, output_name=name)

    return self


def report_spark_lineage(spark_lineage,
                         k,  # type: Kensu
                         result_schema,
                         output_name):
        # similar like GenericComputedInMemDs.report, but using processRun from spark and no Kensu-Py magic
        from kensu.utils.helpers import extract_ksu_ds_schema
        spark_input_names = []
        for input_ds in spark_lineage.inputs:
            ds, schema = extract_ksu_ds_schema(k, input_ds, report=True, register_orig_data=False)
            input_name = ds.pk.location
            spark_input_names.append(input_name)
            k.name_schema_lineage_dict[input_name] = schema.to_guid()

        # report output (if needed)
        result_schema._report()
        lineage_run = link(input_names=spark_input_names, output_name=output_name)
        lineage_run_id = lineage_run.to_guid()
        # ask spark to publish stats given a lineage-run
        # FIXME: do we need to activate/deactivate it depending on config? but kensupy conf or pyspark conf?
        for input_ds in spark_lineage.inputs:
            input_ds.f_publish_stats(lineage_run_id)


def tagInMemWrapper():
    def tagInMemInner(self,  # type: DataFrame
                      name   # type: str
                      ):
        return tagInMem(self, name)
    return tagInMemInner


def tagCreateSparkDataFrameFromPy(
        self,  # type: DataFrame
        name,  # type: str
        input_names=None  # type: list[str]
       ):
    #Get spark session info
    spark = get_spark_from_df(self)
    spark_info = get_process_run_info(spark)
    process = spark_info['process_guid']

    # Creation of an in-mem data source
    location = f'in-memory-data://{process}/{name}'
    in_mem_format = 'in-mem'
    k = KensuProvider().instance()
    ds_pk = DataSourcePK(location=location,
                         physical_location_ref=PhysicalLocationRef(by_pk=k.UNKNOWN_PHYSICAL_LOCATION.pk))
    ds = to_datasource(ds_pk, in_mem_format, location, 'File', name)._report()

    # FIXME: only spark dataframe supported now, i.e. this function must be after spark.createDataFrame()
    fields = get_schema_fields_from_spark_df(spark_df=self)
    schema = Schema(location, pk=SchemaPK(data_source_ref=DataSourceRef(by_guid=ds.to_guid()), fields=fields))._report()
    k.name_schema_lineage_dict[name] = schema.to_guid()

    # report lineage to Kensu between Python inputs and spark.createDataFrame which is not seen by Spark directly
    if input_names:
        link(input_names=input_names, output_name=name)
    else:
        # p.s. if we need mode where in-mem is not reported, would need to get earlier reported lineage
        pass

    # also need to report lineage to Spark JVM,
    # which for now will be simply `name` of this .createDataFrame() in-mem operation
    output_df_tag = f'spark_df_{name}'
    register_manual_lineage(
        spark,
        output_df_tag=output_df_tag,  # p.s. output_df_tag is only used as an ID in Spark agent side, nothing else
        # P.S. or alternatively we could pass over the input_names to Spark,
        #  and then not report any lineage between inputs_name(python) and {spark_dataframe} from within Python
        #  but then the `name` (of this in-memory spark dataframe)  would be not reported to Kensu, so
        #  current solution seems better/more consistent with the rest of this file
        inputs=[
            {
                'path': location,  # FIXME: will DS name be consistently reported between python and Spark/scala?
                'format': in_mem_format,
                'schema': dict([(f.name, f.field_type) for f in fields])
            },
        ]
    )
    output_df = get_df_with_inmem_tag(df=self, df_tag=output_df_tag)
    return output_df


def tagCreateDataFrameWrapper():
    def tagInMemInner(self,  # type: DataFrame
                      name,   # type: str
                      input_names=None
                      ):
        return tagCreateSparkDataFrameFromPy(self, name=name, input_names=input_names)
    return tagInMemInner


def add_tagInMem_operations():
    DataFrame.tagInMem = tagInMemWrapper()
    DataFrame.tagCreateDataFrame = tagCreateDataFrameWrapper()


def create_publish_for_data_source(name, format, location, schema=None, **kwargs):
    k = KensuProvider().instance()
    ds_pk = DataSourcePK(location=location,
                         physical_location_ref=PhysicalLocationRef(by_pk=k.UNKNOWN_PHYSICAL_LOCATION.pk))

    ds = DataSource(name=name,format=format, categories=[f'logical::{name}'],pk=ds_pk)._report()

    if schema is not None:
        fields = schema
    else:
        fields = [FieldDef('unknown','unknown',True)]
    schema = Schema(name, pk=SchemaPK(data_source_ref=DataSourceRef(by_guid=ds.to_guid()),fields=fields))._report()
    k.name_schema_lineage_dict[name] = schema.to_guid()

def create_publish_for_postgres_table(table, location, cur=None):
    format='Postgres Table'
    from kensu.psycopg2.pghelpers import get_table_schema
    try:
        if cur:
            SchemaFields = get_table_schema(cur=cur,table_name=table)
            from kensu.client.models import FieldDef
            schema_ = [FieldDef(f['field_name'],f['field_type'],True) for f in SchemaFields]
        else:
            schema_ = None
    except:
        schema_=None
    create_publish_for_data_source(ds=table, name=location, location=location, format=format, schema=schema_)

def create_publish_for_csv(name, location):
    format='csv'
    import pandas as pd
    from kensu.pandas.extractor import KensuPandasSupport

    try:
        df = pd.read_csv(location)
        schema_ = KensuPandasSupport().extract_schema_fields(df)
    except:
        schema_=None
    create_publish_for_data_source(ds=name, name=name, location=location, format=format, schema=schema_)

def create_publish_for_sklearn_model(name, location):
    format = 'SKLearn'
    k = KensuProvider().instance()
    ds_pk = DataSourcePK(location=location,
                         physical_location_ref=PhysicalLocationRef(by_pk=k.UNKNOWN_PHYSICAL_LOCATION.pk))
    ds = DataSource(name=name,format=format, categories=[f'logical::{name}'],pk=ds_pk)._report()

    fields = [FieldDef('intercept','Numeric',False),FieldDef('coeff','Numeric',False)]
    schema = Schema(name, pk=SchemaPK(data_source_ref=DataSourceRef(by_guid=ds.to_guid()),fields=fields))._report()
    k.name_schema_lineage_dict[name] = schema.to_guid()


def get_schema(name):
   # use the lookup table

   k = KensuProvider().instance()
   if name in k.name_schema_lineage_dict:
       return k.name_schema_lineage_dict[name]
   else:
       logging.info(f'Unable to retrieve Kensu GUID for schema {name}')
       return None


def create_lineage(inputs, output):
    data_flow = [SchemaLineageDependencyDef(from_schema_ref=SchemaRef(by_guid=i),
                                            to_schema_ref=SchemaRef(by_guid=output),
                                            # FIXME: or should we report column_data_dependencies=None ?
                                            column_data_dependencies={'unknown': ['unknown']}) for i in inputs ]
    lineage = ProcessLineage(name=f'Lineage to {str(output)}',  # FIXME: List inputs too?
                             operation_logic='APPEND',
                             pk=ProcessLineagePK(process_ref=ProcessRef(by_guid=get_process_run_info(get_spark_session())['process_guid']),
                                                 data_flow=data_flow))._report()
    return lineage

def link(input_names, output_name):
    # retrieve schemas for inputs
    input_scs = [get_schema(i) for i in input_names]
    # retrieve schemas for outputs
    output_sc = get_schema(output_name)
    # create lineage
    lineage = create_lineage(input_scs,output_sc)._report()
    # create lineage run
    spark_run = get_process_run_info(get_spark_session())['process_run_guid']
    k = KensuProvider().instance()
    lineage_run = LineageRun(LineageRunPK(process_run_ref=ProcessRunRef(by_guid=spark_run),
                                          lineage_ref=ProcessLineageRef(by_guid=lineage.to_guid()),
                                          # Should be int/long
                                          timestamp=int(k.timestamp)))
    print(f'reporting lineage_run for {input_names} -> {output_name}. payload={lineage_run}...')
    lineage_run._report()
    return lineage_run


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
#
# Copyright 2021 Kensu Inc
#

from kensu.utils.kensu_provider import KensuProvider
from google.cloud.bigquery import Table


class BqKensuHelpers:
    @staticmethod
    def table_to_kensu(table: Table):
        kensu = KensuProvider().instance()
        # note: there should be no report here!
        ds = kensu.extractors.extract_data_source(table, kensu.default_physical_location_ref,
                                                  logical_data_source_naming_strategy=kensu.logical_naming)
        sc = kensu.extractors.extract_schema(ds, table)
        return ds, sc




import os
# disable Kensu collector when requested to do so (e.g. inside Apache Spark executor nodes)
if "KSU_DISABLE_PY_COLLECTOR" in os.environ:
    from pandas import *
else:
    import pandas as pd
    from pandas import *

    from .data_frame import DataFrame
    from .data_frame import Series
    from .data_frame import wrap_pandas_reader, wrap_pandas_get_dummies, wrap_merge, wrap_to_datetime, wrap_concat, wrap_json_normalize, wrap_all_function
    from .extractor import KensuPandasSupport

    functions=dir(pd)

    if hasattr(pd, "read_clipboard"):
        read_clipboard = wrap_pandas_reader(pd.read_clipboard)
        functions.remove("read_clipboard")
    if hasattr(pd, "read_csv"):
        read_csv = wrap_pandas_reader(pd.read_csv)
        functions.remove("read_csv")
    if hasattr(pd, "read_excel"):
        read_excel = wrap_pandas_reader(pd.read_excel)
        functions.remove("read_excel")
    if hasattr(pd, "read_feather"):
        read_feather = wrap_pandas_reader(pd.read_feather)
        functions.remove("read_feather")
    if hasattr(pd, "read_fwf"):
        read_fwf = wrap_pandas_reader(pd.read_fwf)
        functions.remove("read_fwf")
    if hasattr(pd, "read_gbq"):
        read_gbq = wrap_pandas_reader(pd.read_gbq)
        functions.remove("read_gbq")
    if hasattr(pd, "read_hdf"):
        read_hdf = wrap_pandas_reader(pd.read_hdf)
        functions.remove("read_hdf")
    if hasattr(pd, "read_html"):
        read_html = wrap_pandas_reader(pd.read_html)
        functions.remove("read_html")
    if hasattr(pd, "read_json"):
        read_json = wrap_pandas_reader(pd.read_json)
        functions.remove("read_json")
    if hasattr(pd, "read_parquet"):
        read_parquet = wrap_pandas_reader(pd.read_parquet)
        functions.remove("read_parquet")
    if hasattr(pd, "read_pickle"):
        read_pickle = wrap_pandas_reader(pd.read_pickle)
        functions.remove("read_pickle")
    if hasattr(pd, "read_sas"):
        read_sas = wrap_pandas_reader(pd.read_sas)
        functions.remove("read_sas")
    if hasattr(pd, "read_sql"):
        read_sql = wrap_pandas_reader(pd.read_sql)
        functions.remove("read_sql")
    if hasattr(pd, "read_sql_query"):
        read_sql_query = wrap_pandas_reader(pd.read_sql_query)
        functions.remove("read_sql_query")
    if hasattr(pd, "read_sql_table"):
        read_sql_table = wrap_pandas_reader(pd.read_sql_table)
        functions.remove("read_sql_table")
    if hasattr(pd, "get_dummies"):
        get_dummies = wrap_pandas_get_dummies(pd.get_dummies)
        functions.remove("get_dummies")
    if hasattr(pd, "merge"):
        merge = wrap_merge(pd.merge)
        functions.remove("merge")
    if hasattr(pd, 'to_datetime'):
        to_datetime = wrap_to_datetime(pd.to_datetime)
        functions.remove("to_datetime")
    if hasattr(pd, 'concat'):
        concat = wrap_concat(pd.concat)
        functions.remove("concat")
    if hasattr(pd, 'json_normalize'):
        json_normalize = wrap_json_normalize(pd.json_normalize)
        functions.remove("json_normalize")



from kensu.utils.dsl.ended_builder_element import EndedBuilderElement


class AddDepsBuilderElement(object):
    def __init__(self, builder):
        self.builder = builder

    def link(self, output_field_name, input_fields_name_array, type):
        c = self.builder.control
        if type == "data": c = self.builder.data
        c[output_field_name] = input_fields_name_array
        return self

    def filter_io(self, outs, ins):
        if outs is None:
            kept_outs = self.builder.output_fields_name
        else:
            kept_outs = [n for n in self.builder.output_fields_name if n in outs]

        if ins is None:
            kept_ins = self.builder.input_fields_name
        else:
            kept_ins = [n for n in self.builder.input_fields_name if n in ins]

        return (kept_outs, kept_ins)

    def with_strategy(self, mapping_strategy, type="data"):
        return mapping_strategy.map(self, type)

    def direct(self, outs=None, ins=None, type="data"):
        c = self.builder.control
        if type == "data": c = self.builder.data

        (kept_outs, kept_ins) = self.filter_io(outs, ins)
        for f in kept_outs:
            if f in kept_ins:
                c[f] = [f]
        return self

    def full(self, outs=None, ins=None, type="data"):
        c = self.builder.control
        if type == "data": c = self.builder.data

        (kept_outs, kept_ins) = self.filter_io(outs, ins)

        for f in kept_outs:
            c[f] = kept_ins
        return self

    @property
    def e(self):
        return self.end_dependency()

    def end_dependency(self):
        self.builder.lineage_builder.add_deps(EndedBuilderElement(self.builder))
        return self.builder.lineage_builder


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
#
# Copyright 2021 Kensu Inc
#
import logging
import traceback

from kensu.google.cloud.bigquery.job.bq_helpers import BqKensuHelpers
from kensu.google.cloud.bigquery.job.bigquery_stats import compute_bigquery_stats
from kensu.utils.dsl.extractors.external_lineage_dtos import KensuDatasourceAndSchema, GenericComputedInMemDs, \
    ExtDependencyEntry
import google.cloud.bigquery as bq
from kensu.utils.kensu_provider import KensuProvider

logger = logging.getLogger(__name__)


class BqRemoteParser:

    @staticmethod
    def get_headers():
        k = KensuProvider().instance()
        try:
            if k.bigquery_headers:
                headers = k.bigquery_headers
        except:
            headers = None
        return headers

    @staticmethod
    def parse(kensu, client: bq.Client, query: str, db_metadata, table_id_to_bqtable) -> GenericComputedInMemDs:
        ## POST REQUEST to /lineage-and-stats-criterions
        req = {"sql": query, "metadata": db_metadata}
        url = kensu.kensu_sql_parser_url
        logger.debug("sending request to SQL parsing service url={} request={}".format(url, str(req)))
        import requests
        def convert(fieldtype):
            if fieldtype in ["INT64", "INT", "SMALLINT", "INTEGER", "BIGINT", "TINYINT", "BYTEINT", "BIGNUMERIC","NUMERIC"]:
                return "INTEGER"
            else:
                return fieldtype
        for table in db_metadata['tables']:
            for field in table['schema']['fields'] :
                field['type'] = convert(field['type'])

        #TODO Support UNNEST in queries  'SELECT e.key FROM `psyched-freedom-306508.my_dataset.sample`, UNNEST(user_properties) as e '
        lineage_resp = requests.post(url + "/lineage-and-stats-criterions", json=req, headers = BqRemoteParser.get_headers())
        logger.debug("lineage_resp:" + str(lineage_resp))
        logger.debug("lineage_resp_body:" + str(lineage_resp.text))
        parsed_resp = lineage_resp.json()
        lineage_info = parsed_resp['lineage']
        stats_info = parsed_resp['stats']
        lineage = list([BqRemoteParser.convert_lineage_entry(
                lineage_entry,
                kensu=kensu,
                client=client,
                table_id_to_bqtable=table_id_to_bqtable,
                stats_info=stats_info
            ) for lineage_entry in lineage_info])
        converted_lineage = GenericComputedInMemDs(lineage=lineage)
        logger.debug('converted_lineage:' + str(converted_lineage))
        return converted_lineage

    @staticmethod
    def convert_lineage_entry(lineage_entry, kensu, client: bq.Client, table_id_to_bqtable, stats_info):
        table_id = lineage_entry['table']
        logger.debug('table_id = {}, table_id_to_bqtable.keys={}'.format(table_id, str(table_id_to_bqtable)))
        bq_table = table_id_to_bqtable.get(table_id)
        stats_values = {}
        ds_name = None
        if bq_table is not None:
            ds, sc = BqKensuHelpers.table_to_kensu(bq_table)
            ds_path = ds.pk.location
            ds_name = ds.name
            sc = [(f.name, f.field_type) for f in sc.pk.fields]
            table_stats_info = stats_info.get(table_id, {})
            stats_aggs = table_stats_info.get('stats')
            stats_filters = table_stats_info.get('input_filters')
            bg_table_ref = bq_table.reference
            # note: making stats computation lazy in a f_get_stats lambda seem to behave very weirdly...
            # so stats are computed eagerly now
            if kensu.compute_stats:
                try:
                    stats_values = compute_bigquery_stats(
                        table_ref=bg_table_ref,
                        table=bq_table,
                        client=client,
                        stats_aggs=stats_aggs,
                        input_filters=stats_filters)
                except Exception as e:
                    stats_values = None
                    logger.warning(f"Unable to compute the stats: {e}")

            logger.debug(
                f'table_id {table_id} (table.ref={bg_table_ref}, ds_path: {ds_path}) got input_filters: {stats_filters} & stat_aggs:{str(stats_aggs)}')
        else:
            logger.warning('table_id={} as reported by remote service was not found in table_id cache: {}'.format(
                table_id, str(table_id_to_bqtable.keys())))
            sc = None
            ds_path = 'bigquery:/' + table_id  # FIXME: add proper BQ prefix, and extract a shared helper
        input = KensuDatasourceAndSchema.for_path_with_opt_schema(
            kensu,
            ds_path=ds_path,
            ds_name=ds_name,
            format='BigQuery table',
            categories=None,
            maybe_schema=sc,
            f_get_stats=lambda: stats_values
        )
        return ExtDependencyEntry(
            input_ds=input,
            lineage=lineage_entry['mapping']
        )


import json as js
from json import *

from kensu.boto3 import ksu_dict
from kensu.botocore.response import ksu_bytes
from kensu.client import DataSourcePK, DataSource, FieldDef, SchemaPK, Schema
from kensu.itertools import kensu_list
from kensu.utils.dsl.extractors.external_lineage_dtos import KensuDatasourceAndSchema
from kensu.utils.helpers import logical_naming_batch, extract_short_json_schema
from kensu.utils.kensu_provider import KensuProvider


def wrap_loads(method):
    def wrapper(*args, **kwargs):

        result = method(*args, **kwargs)

        kensu = KensuProvider().instance()

        data = args[0]
        if hasattr(data, 'ksu_metadata'):
            if isinstance(result, dict):
                result = ksu_dict(result)
                result.ksu_metadata = data.ksu_metadata
            elif isinstance(result, list):
                result = kensu_list(result)
                result.ksu_metadata = data.ksu_metadata
            if isinstance(data, ksu_bytes):
                result.ksu_metadata = data.ksu_metadata

                origin_location = result.ksu_metadata['origin_location']
                origin_name = result.ksu_metadata['origin_name']

                if kensu.logical_naming == 'ReplaceNumbers':
                    logical = logical_naming_batch(origin_name)
                else:
                    logical = origin_name

                # FIXME we should put this in extractors
                result_pk = DataSourcePK(location=origin_location,
                                         physical_location_ref=kensu.default_physical_location_ref)
                result_ds = DataSource(name=origin_name, categories=['logical::' + logical],
                                       format=origin_name.split('.')[-1],
                                       pk=result_pk)._report()

                short_result_sc = extract_short_json_schema(result, result_ds)._report()

                kensu.real_schema_df[short_result_sc.to_guid()] = None

                kensu.add_input_ref(KensuDatasourceAndSchema(ksu_ds=result_ds, ksu_schema=short_result_sc))

        return result

    wrapper.__doc__ = method.__doc__
    return wrapper

loads = wrap_loads(js.loads)


# coding: utf-8

import setuptools
from setuptools import setup, find_packages
import os

NAME = "kensu"


BUILD_FLAVOR = os.environ["BUILD_FLAVOR"] if "BUILD_FLAVOR" in os.environ else ""
BUILD_NUMBER = os.environ["BUILD_NUMBER"] if "BUILD_NUMBER" in os.environ else ""
# https://semver.org/
VERSION = "2.5.1" + BUILD_FLAVOR + BUILD_NUMBER



# To install the library, run the following
#
# python setup.py install
#
# prerequisite: setuptools
# http://pypi.python.org/pypi/setuptools

def get_extra_requires(path, add_all=True, add_all_but_test=True, add_no_extra_deps=True):
    import re
    from collections import defaultdict

    with open(path) as fp:
        extra_deps = defaultdict(set)
        for k in fp:
            if k.strip() and not k.startswith('#'):
                tags = set()
                if ':' in k:
                    k, v = k.split(':')
                    tags.update(vv.strip() for vv in v.split(','))
                tags.add(re.split('[<=>]', k)[0])
                for t in tags:
                    extra_deps[t].add(k)

        # add tag `all` at the end
        if add_all:
            extra_deps['all'] = set(vv for v in extra_deps.values() for vv in v)

        # add tag `all-but-test` at the end
        if add_all_but_test:
            extra_deps['all-but-test'] = set(vv for v in extra_deps.values() for vv in v if vv != "test")

        if add_no_extra_deps:
            extra_deps['no-extra-deps'] = set()


    print("Collected the following dependencies from " + path + ":")
    print(extra_deps)

    return extra_deps


def get_install_requires(path):
    with open(path) as fp:
        deps = []
        for l in fp:
            if l.strip() and not l.startswith('#'):
                deps.append(l)
        return deps

setup(
    name=NAME,
    version=VERSION,
    description="",
    author_email="",
    url="",
    keywords=["DODD", "Data Observability Driven Development", "Data Observability", "Analytics Observability"],
    packages=[
        package
        for package in setuptools.PEP420PackageFinder.find()
        if package.startswith("kensu")
    ],
    install_requires=get_install_requires('common-requirements.txt'),
    extras_require=get_extra_requires('extra.requirements'),
    platforms="Posix; MacOS X; Windows",
    include_package_data=True,
    long_description="""\
    DODD Python Agent: enable Data Observability Driven Development in your Python script\
    """
)
